@@ -337,12 +337,12 @@ static int nfs_verify_lock_option(struct mount_options *options)
 	return 1;
 }
 
-static int nfs_append_sloppy_option(struct mount_options *options)
+static int nfs_insert_sloppy_option(struct mount_options *options)
 {
 	if (!sloppy || linux_version_code() < MAKE_VERSION(2, 6, 27))
 		return 1;
 
-	if (po_append(options, "sloppy") == PO_FAILED)
+	if (po_insert(options, "sloppy") == PO_FAILED)
 		return 0;
 	return 1;
 }
@@ -425,7 +425,7 @@ static int nfs_validate_options(struct nfsmount_info *mi)
 	if (!nfs_set_version(mi))
 		return 0;
 
-	if (!nfs_append_sloppy_option(mi->options))
+	if (!nfs_insert_sloppy_option(mi->options))
 		return 0;
 
 	return 1;
